extern "C" {
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
}

#include "error.hh"

basic_exception::basic_exception(const char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    vsnprintf(&msg_[0], sizeof(msg_), fmt, ap);
    va_end(ap);
}

basic_exception::basic_exception(const basic_exception &o)
    : std::exception(o)
{
    memcpy(&msg_[0], &o.msg_[0], sizeof(msg_));
}

basic_exception::~basic_exception() throw ()
{
}

void
basic_exception::set_msg(const char *msg)
{
    snprintf(&msg_[0], sizeof(msg_), "%s", msg);
}

void
basic_exception::print_trace() const
{
    fprintf(stderr, "basic_exception::print_where(): backtraces disabled\n");
}

error::error(int r, const char *fmt, ...) : err_(r)
{
    char buf[256];
    int used = 0;
    va_list ap;

    va_start(ap, fmt);
    used += vsnprintf(&buf[0], sizeof(buf), fmt, ap);
    va_end(ap);

    used += snprintf(&buf[used], sizeof(buf) - used, ": %s", strerror(r));

    buf[sizeof(buf) - 1] = '\0';
    set_msg(&buf[0]);
}
