#include "locks.h"
#include <stdio.h>
#include <stdlib.h>
#include "bench.h"

struct atomic_uint64_t {
    uint64_t v;
};

#define MAX_LOCKS 3000

typedef union {
    struct qnode v;
    char __pad[JOS_CLINE];
} aligned_qnode;

static __thread volatile aligned_qnode I[MAX_LOCKS];
static struct atomic_uint64_t lock_used[MAX_LOCKS];

static __inline__ uint64_t
cmp_and_swap_atomic(struct atomic_uint64_t *L, uint64_t cmpval, uint64_t newval)
{
    uint64_t out;
    __asm__ volatile(
    		"lock; cmpxchgq %2, %1"
		: "=a" (out), "+m" (L->v)
		: "q" (newval), "0"(cmpval)
		: "cc");
    return out == cmpval;
}

static __inline__ uint64_t
fetch_and_store(mcslock *L, uint64_t val) 
{
    uint64_t out;
    __asm__ volatile(
    		"movq %2, %%r10\n\t"
		"lock; xchgq %1, %%r10\n\t"
		"movq %%r10, %0"
		: "=a" (out), "+m" (L->v)
		: "q" (val)
		: "%r10", "memory", "cc");
    return out;
}

static __inline__ uint64_t
cmp_and_swap(mcslock *L, uint64_t cmpval, uint64_t newval)
{
    uint64_t out;
    __asm__ volatile(
    		"lock; cmpxchgq %2, %1"
		: "=a" (out), "+m" (L->v)
		: "q" (newval), "0"(cmpval)
		: "cc");
    return out == cmpval;
}

void 
mcs_init(mcslock *L)
{
    L->v = NULL;
    for (int i = 0; i < MAX_LOCKS; i++)
        if (!lock_used[i].v)
	    if (cmp_and_swap_atomic(&lock_used[i], 0, 1)) {
	        L->lock_idx = i;
	        return;
	    }
    printf("MCS lock exhausted\n");
    exit(0);
}

void
mcs_lock(mcslock *L)
{
    volatile struct qnode *mynode = &I[L->lock_idx].v;
    mynode->next = NULL;
    struct qnode *predecessor = (struct qnode *)fetch_and_store(L, (uint64_t)mynode);
    if (predecessor) {
	mynode->locked = 1;
	predecessor->next = mynode;
	while (mynode->locked)
	    nop_pause();
    }
}

void
mcs_unlock(mcslock *L)
{
    volatile struct qnode *mynode = &I[L->lock_idx].v;
    if (!mynode->next) {
        if (cmp_and_swap(L, (uint64_t)mynode, 0)) {
	    return;
	}
	while (!mynode->next)
	    nop_pause();
    }
    ((struct qnode *)mynode->next)->locked = 0;
}
