#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <numa.h>

#include "lib/bench.h"
#include "lib/pad.h"

enum { buf_size = 1024 * 1024 * 1024 };
enum { buf_reads = buf_size / 4096 };

struct memnet_conf {
    uint32_t dst;
    uint32_t src;
    void *   va;
};

static struct memnet_conf one_hop[] = { 
    { 4, 0, 0 },
};

static struct memnet_conf two_hops[] = { 
    { 12, 0, 0 },
};

static struct memnet_conf one_core[] = { 
    { 0, 0, 0 },
};

static struct memnet_conf two_core[] = { 
    { 0, 0, 0 },
    { 1, 0, 0 },
};

static struct memnet_conf three_core[] = { 
    { 0, 0, 0 },
    { 1, 0, 0 },
    { 2, 0, 0 },
};

static struct memnet_conf four_core[] = { 
    { 0, 0, 0 },
    { 1, 0, 0 },
    { 2, 0, 0 },
    { 3, 0, 0 },
};

static struct memnet_conf two_node[] = { 
    { 0, 0, 0 },
    { 1, 0, 0 },
    { 2, 0, 0 },

    { 4, 4, 0 },
    { 5, 4, 0 },
    { 6, 4, 0 },
};

static struct memnet_conf three_node[] = { 
    { 0, 0, 0 },
    { 1, 0, 0 },
    { 2, 0, 0 },

    { 4, 4, 0 },
    { 5, 4, 0 },
    { 6, 4, 0 },

    { 8, 8, 0 },
    { 9, 8, 0 },
    { 10, 8, 0 },
};

static struct memnet_conf four_node[] = { 
    { 0, 0, 0 },
    { 1, 0, 0 },
    { 2, 0, 0 },

    { 4, 4, 0 },
    { 5, 4, 0 },
    { 6, 4, 0 },

    { 8, 8, 0 },
    { 9, 8, 0 },
    { 10, 8, 0 },

    { 12, 12, 0 },
    { 13, 12, 0 },
    { 14, 12, 0 },
};

static struct memnet_conf two_far_xfire[] = { 
    { 0, 12, 0 },
    { 1, 12, 0 },
    { 2, 12, 0 },

    { 12, 0, 0 },
    { 13, 0, 0 },
    { 14, 0, 0 },
};

static struct memnet_conf two_near_xfire[] = { 
    { 0, 4, 0 },
    { 1, 4, 0 },
    { 2, 4, 0 },
    
    { 4, 0, 0 },
    { 5, 0, 0 },
    { 6, 0, 0 },
};

static struct memnet_conf three_xfire[] = { 
    { 0, 4, 0 },
    { 1, 8, 0 },
    { 2, 12, 0 },

    { 4, 0, 0 },
    { 5, 8, 0 },
    { 6, 12, 0 },

    { 8, 0, 0 },
    { 9, 4, 0 },
    { 10, 12, 0 },
};

static struct memnet_conf four_xfire[] = { 
    { 0, 4, 0 },
    { 1, 8, 0 },
    { 2, 12, 0 },

    { 4, 0, 0 },
    { 5, 8, 0 },
    { 6, 12, 0 },

    { 8, 0, 0 },
    { 9, 4, 0 },
    { 10, 12, 0 },

    { 12, 0, 0 },
    { 13, 4, 0 },
    { 14, 8, 0 },
};

static struct memnet_conf one_by_four[] = {
    { 0, 0, 0},

    { 4, 4, 0},

    { 8, 8, 0},
    
    { 12, 12, 0},
};

static struct memnet_conf two_by_four[] = {
    { 0, 0, 0},
    { 1, 1, 0},

    { 4, 4, 0},
    { 5, 5, 0},

    { 8, 8, 0},
    { 9, 9, 0},
    
    { 12, 12, 0},
    { 13, 13, 0},
};

static struct memnet_conf three_by_four[] = {
    { 0, 0, 0},
    { 1, 1, 0},
    { 2, 2, 0},

    { 4, 4, 0},
    { 5, 5, 0},
    { 6, 6, 0},

    { 8, 8, 0},
    { 9, 9, 0},
    { 10, 10, 0},
    
    { 12, 12, 0},
    { 13, 13, 0},
    { 14, 14, 0},
};

static struct memnet_conf four_by_four[] = {
    { 0, 0, 0},
    { 1, 1, 0},
    { 2, 2, 0},
    { 3, 3, 0},

    { 4, 4, 0},
    { 5, 5, 0},
    { 6, 6, 0},
    { 7, 7, 0},

    { 8, 8, 0},
    { 9, 9, 0},
    { 10, 10, 0},
    { 11, 11, 0},
    
    { 12, 12, 0},
    { 13, 13, 0},
    { 14, 14, 0},
    { 15, 15, 0},
};

struct bench_stats {
    struct memnet_conf *conf;
    uint32_t nconf;

    volatile uint64_t signal;
    uint32_t master;

    PAD_TYPE(uint64_t, JOS_CLINE) count[JOS_NCPU];
};

static volatile struct bench_stats *stats = NULL;

enum { two_unsynced = 0, two_synced, all_start, all_end};

void read_page(volatile void *va);

static void
xfire_read(void *buf, uint32_t cid)
{
    void *src = buf;
    void *src_end = src + buf_size;
    char is_master = stats->master == cid;
    
    while (stats->signal == all_start) {
	read_page(src);

	stats->count[cid].v++;
	if (is_master && stats->count[cid].v == buf_reads)
	    stats->signal = all_end;

       	src = src + 4096;
	if (src >= src_end)
	    src = buf;
    }
}

static void *
xfire_core(void *arg)
{
    int conf_index = (int)(uint64_t)arg;
    int cid = stats->conf[conf_index].dst;
    affinity_set(cid);

    stats->signal = two_synced;

    while (stats->signal != all_start)
	nop_pause();
   
    xfire_read(stats->conf[conf_index].va, cid);
    return NULL;
}

static void
prefault(char *buf, uint64_t size)
{
    for (uint64_t i = 0; i < size; i += 4096)
        buf[i] = 0;
}

static void __attribute__((unused))
do_bench(const char *name, struct memnet_conf *conf, uint32_t nconf)
{
    assert(nconf);
    memset((void *)stats, 0, sizeof(*stats));

    stats->conf = conf;
    stats->nconf = nconf;
    stats->master = conf[0].dst;
    stats->signal = two_unsynced;

    void *src = 0;
    for (uint32_t i = 0; i < nconf; i++) {
	int node = conf[i].src / (get_core_count() / (numa_max_node() + 1));
        errno_check(conf[i].va = numa_alloc_onnode(buf_size, node));
	prefault(conf[i].va, buf_size);
	if (conf[i].dst) {
	    pthread_start(xfire_core, i);
	    while (stats->signal != two_synced)
		nop_pause();
	    stats->signal = two_unsynced;
	}
	else {
	    src = conf[i].va;
	}
    }

    stats->signal = all_start;

    uint64_t s = read_tsc();
    if (src)
	xfire_read(src, 0);
    else
	while (stats->signal != all_end)
	    nop_pause();
    uint64_t e = read_tsc();
    
    uint64_t bytes = 0;
    for (uint32_t i = 0; i < JOS_NCPU; i++)
	bytes += (stats->count[i].v * 4096);
    
    printf("%-12s bytes %ld cycle %ld bytes/cycle %.2f\n", name, bytes, e - s, bytes / (double)(e - s));

    for (uint32_t i = 0; i < nconf; i++) {
	if (conf[i].va) {
	    numa_free(conf[i].va, buf_size);
	    conf[i].va = 0;
	}
    }
    sleep(1);
}

int
main(int argc, char **argv)
{
    affinity_set(0);
    stats = (volatile struct bench_stats *)malloc(sizeof(*stats));
    errno_check(numa_available());

    do_bench("1 hop", one_hop, array_size(one_hop));
    do_bench("2 hops", two_hops, array_size(two_hops));

    do_bench("1 core", one_core, array_size(one_core));
    do_bench("2 core", two_core, array_size(two_core));
    do_bench("3 core", three_core, array_size(three_core));
    do_bench("4 core", four_core, array_size(four_core));
    do_bench("2 node", two_node, array_size(two_node));    
    do_bench("3 node", three_node, array_size(three_node));
    do_bench("4 node", four_node, array_size(four_node));
    do_bench("2 near xfire", two_near_xfire, array_size(two_near_xfire));
    do_bench("2 far xfire", two_far_xfire, array_size(two_far_xfire));
    do_bench("3 xfire", three_xfire, array_size(three_xfire));
    do_bench("4 xfire", four_xfire, array_size(four_xfire));
    do_bench("1 x 4", one_by_four, array_size(one_by_four));
    do_bench("2 x 4", two_by_four, array_size(two_by_four));
    do_bench("3 x 4", three_by_four, array_size(three_by_four));
    do_bench("4 x 4", four_by_four, array_size(four_by_four));
    return 0;
}
