#include <sys/mman.h>
#include <asm/page.h>
#include <stdio.h>

#include "lib/bench.h"

static struct {
    volatile int start;
    union {
	struct {
	    volatile int ready;
	    volatile uint64_t cycle;
	};
	char pad[JOS_CLINE];
    } cpu[JOS_NCPU] __attribute__((aligned(JOS_CLINE)));
} *sync_state;

static int nbytes;

static uint32_t affinity[] = {
    0, 4, 8, 12,
    1, 5, 9, 13,
    2, 6, 10, 14,
    3, 7, 11, 15,
};

static void *
worker(void *x)
{
    int i = (uintptr_t)x;
    affinity_set(affinity[i]);
    char *buf = malloc(nbytes);
    if (!buf)
	eprint("malloc failed\n");
    char *end = buf + nbytes;

    sync_state->cpu[i].ready = 1;
    if (i)
	while (!sync_state->start)
	    nop_pause();
    else
	sync_state->start = 1;

    uint64_t s = read_tsc();
    for (; buf != end; buf += PAGE_SIZE)
	*buf = 1;

    sync_state->cpu[i].cycle = read_tsc() - s;
    return 0;
}

int
main(int ac, char **av)
{
    affinity_set(affinity[0]);
    if (ac < 3) {
	fprintf(stderr, "usage: %s num-cores num-mbytes use-threads\n", av[0]);
	return -1;
    }

    int ncores = atoi(av[1]);
    nbytes = atoi(av[2]) * 1024 * 1024;
    int threads = atoi(av[3]);

    sync_state = mmap(0, sizeof(*sync_state), 
		      PROT_READ | PROT_WRITE, 
		      MAP_SHARED | MAP_ANONYMOUS, 0, 0);
    if (sync_state == MAP_FAILED)
	eprint("mmap failed\n");
    memset(sync_state, 0, sizeof(*sync_state));
    
    for (int i = 1; i < ncores; i++) {
	if (threads)
	    pthread_start(worker, i);
	else {
	    pid_t p = fork();
	    if (p < 0)
		eprint("fork failed\n");
	    if (!p) {
		worker((void *)(intptr_t)i);
		return 0;
	    }
	}

	while (!sync_state->cpu[i].ready)
	    nop_pause();
    }

    worker((void *)(intptr_t)0);

    uint64_t tot = 0;
    uint64_t max = 0;
    for (int i = 0; i < ncores; i++) {
	while (!sync_state->cpu[i].cycle)
	    nop_pause();
	tot += sync_state->cpu[i].cycle;
	if (sync_state->cpu[i].cycle > max)
	    max = sync_state->cpu[i].cycle;
    }

    uint64_t ave_pf = (tot / ncores) / (nbytes / PAGE_SIZE);
    uint64_t max_pf = max / (nbytes / PAGE_SIZE);
    printf("ave cycles/pf %ld, max cycles/pf %ld\n", ave_pf, max_pf);
}
