#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <assert.h>
#include <stdint.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/user.h>
#include <time.h>
#include <fcntl.h>
#include <signal.h>
#include <sched.h>

#include <string.h>
#include <pthread.h>

#include "lib/bench.h"

enum { num_bufs = 25600 };

static char *page_buf = NULL;
static int ncores = 0;

static struct {
    uint64_t start;
    union {
	volatile int signal;
	char pad[JOS_CLINE];
    } cpu[JOS_NCPU] __attribute__((aligned(JOS_CLINE)));
    volatile int end;
    volatile int worker_ready;
} *sync_state;

static void __attribute__((noreturn))
xperror(const char *msg)
{
    perror(msg);
    exit(-1);
}

void *
potato_test(void *a)
{
    int core = (int)(intptr_t)a;
    affinity_set(core);

    if (core) {
        sync_state->worker_ready = 1;
	while (sync_state->cpu[core].signal == 0)
	    nop_pause();
    }
    else
	sync_state->start = read_tsc();

    for (int i = 0; i < num_bufs; i++)
	page_buf[i * 4096] = core;
    
    sync_state->cpu[(core + 1) % ncores].signal = 1;

    if (core)
	while (!sync_state->end)
	    nop_pause();
    
    return 0;
}

int
main(int ac, char **av)
{
    affinity_set(0);
    if (ac < 3) {
	fprintf(stderr, "usage: %s sep_address num-cores\n", av[0]);
	return -1;
    }
    int sep_address = atoi(av[1]);
    ncores = atoi(av[2]);
    void *buf = mmap(0, (1 + num_bufs) * PAGE_SIZE, 
	       PROT_READ | PROT_WRITE, 
	       MAP_SHARED | MAP_ANONYMOUS, 0, 0);
    assert(buf != (void *)-1);
    sync_state = buf;
    assert(sizeof(*sync_state) < PAGE_SIZE);
    page_buf = buf + PAGE_SIZE;
    memset(sync_state, 0, sizeof(*sync_state));

    for (int i = 1; i < ncores; i++) {
        sync_state->worker_ready = 0;
	if (!sep_address) {
	    pthread_start(potato_test, (uintptr_t)i);
	} else {
	    pid_t p = fork();
	    assert(p >= 0);
	    if (!p) {
		potato_test((void *)(intptr_t)i);
		return 0;
	    }
	}
	while (!sync_state->worker_ready)
	    nop_pause();
    }

    potato_test((void *)0);
    while (!sync_state->cpu[0].signal)
	nop_pause();
    
    uint64_t end = read_tsc();
    uint64_t usec = (end - sync_state->start) * 1000000 / get_cpu_freq();
    printf("%d\t%"PRIu64"\t\n", ncores, usec);

    sync_state->end = 1;
    return 0;
}
